#!/usr/bin/env python3

"""The Tracery commandline interface."""

import argparse
import json

import tracery
from tracery.modifiers import base_english


def main():
    parser = argparse.ArgumentParser(
        description="Output an example line from a JSON Tracery grammar.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("json", help="Input JSON file")
    parser.add_argument(
        "number", type=int, default=1, nargs="?", help="Number of lines to generate"
    )
    args = parser.parse_args()

    with open(args.json) as data_file:
        rules = json.load(data_file)

    grammar = tracery.Grammar(rules)
    grammar.add_modifiers(base_english)

    for _ in range(args.number):
        print(grammar.flatten("#origin#"))


if __name__ == "__main__":
    main()



"""PyTracery

PyTracery is a Python-port of the JavaScript library, by GalaxyKate, that uses grammars
to generate surprising new text.

"""

from __future__ import annotations

import enum
import random
import re
from argparse import Action
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Protocol, Tuple, Union


class NodeType(enum.IntEnum):
    Raw = -1
    """Needs parsing."""
    PlainText = 0
    """Plain text."""
    Tag = 1
    """('#symbol.mod.mod2.mod3#' or '#[pushTarget:pushRule]symbol.mod')"""
    Action = 2
    """('[pushTarget:pushRule]', '[pushTarget:POP]', more in the future)"""


class Node:
    """A single node in the grammar expansion."""

    __slots__ = (
        "errors",
        "parent",
        "grammar",
        "child_index",
        "node_type",
        "depth",
        "is_expanded",
        "raw",
        "children",
        "finished_text",
        "modifiers",
    )

    errors: List[str]
    """Errors accumulated when evaluating a node."""

    parent: Optional[Node]
    """The parent of a node."""

    grammar: Grammar
    """The grammar a node belongs to."""

    child_index: int
    """The index of a node in it's parent's child list."""

    node_type: NodeType
    """The type of the node."""

    is_expanded: bool
    """Has a node been expanded."""

    depth: int
    """The depth of a node in the tree."""

    raw: str
    """The raw text for of a node."""

    finished_text: str
    """The final expanded text."""

    children: List[Node]
    """The children of a node."""

    modifiers: List[str]
    """Names of modifiers to apply to this node."""

    def __init__(
        self,
        parent: Optional[Node],
        grammar: Grammar,
        child_index: int,
        settings: _ParsedRuleSection,
    ) -> None:
        self.children: List[Node] = []
        self.errors = []
        self.finished_text = ""
        self.grammar = grammar
        self.child_index = child_index
        self.parent = parent
        self.modifiers = []

        if parent is not None:
            self.depth = parent.depth + 1
        else:
            self.depth = 0

        self.raw = settings.raw
        self.node_type = settings.node_type
        self.is_expanded = False

    def expand_children(
        self, child_rule: Optional[str], prevent_recursion: bool = False
    ):
        self.finished_text = ""
        self.children = []

        if child_rule is not None:
            sections, errors = parse(child_rule)
            self.errors.extend(errors)
            for i, section in enumerate(sections):
                node = Node(self, self.grammar, i, section)
                self.children.append(node)
                if not prevent_recursion:
                    node.expand(prevent_recursion)
                self.finished_text += node.finished_text
        else:
            self.errors.append("No child rule provided, can't expand children.")

    def expand(self, prevent_recursion: bool = False) -> None:
        if not self.is_expanded:
            self.is_expanded = True

            if self.node_type == NodeType.Raw:
                self.expand_children(self.raw, prevent_recursion)

            elif self.node_type == NodeType.PlainText:
                self.finished_text = self.raw

            elif self.node_type == NodeType.Tag:
                pre_actions: List[NodeAction] = []
                post_actions: List[NodeAction] = []

                parsed = parse_tag(self.raw)

                for action in parsed.pre_actions:
                    pre_actions.append(NodeAction(self, action.raw))

                for action in pre_actions:
                    if action.action_type == ActionType.Push:
                        undo_action = action.create_undo()
                        if undo_action is not None:
                            post_actions.append(undo_action)

                for action in pre_actions:
                    action.activate()

                self.finished_text = self.raw

                selected_rule = self.grammar.select_rule(
                    parsed.symbol, self, self.errors
                )
                self.expand_children(selected_rule, prevent_recursion)

                # apply modifiers
                for mod_name in parsed.modifiers:
                    mod_params: List[str] = []

                    if mod_name.find("(") > 0:
                        regexp = re.compile(r"\(([^)]+)\)")
                        matches = regexp.findall(mod_name)
                        if len(matches) > 0:
                            mod_params = matches[0].split(",")
                            mod_name = mod_name[: mod_name.find("(")]

                    mod = self.grammar.modifiers.get(mod_name, None)
                    if mod is None:
                        self.errors.append("Missing modifier " + mod_name)
                        self.finished_text += "((." + mod_name + "))"
                    else:
                        self.finished_text = mod(self.finished_text, *mod_params)

            elif self.node_type == NodeType.Action:
                action = NodeAction(self, self.raw)
                action.activate()
                self.finished_text = ""

    def clear_escape_chars(self):
        """Remove all escape characters."""
        self.finished_text = (
            self.finished_text.replace("\\\\", "DOUBLEBACKSLASH")
            .replace("\\", "")
            .replace("DOUBLEBACKSLASH", "\\")
        )


class ActionType(enum.IntEnum):
    Push = 0
    Pop = 1
    Flatten = 2


class NodeAction:
    __slots__ = (
        "node",
        "target",
        "rule",
        "action_type",
    )

    node: Node
    """The node to perform an action on."""

    target: str
    """A section of text to apply the action to."""

    action_type: ActionType
    """The type of action performed."""

    def __init__(self, node: Node, raw: str) -> None:
        self.node = node
        sections = raw.split(":")
        self.target = sections[0]

        if len(sections) == 1:
            self.action_type = ActionType.Flatten
        else:
            self.rule = sections[1]
            if self.rule == "POP":
                self.action_type = ActionType.Pop
            else:
                self.action_type = ActionType.Push

    def create_undo(self) -> Optional[NodeAction]:
        if self.action_type == ActionType.Push:
            return NodeAction(self.node, self.target + ":POP")
        return None

    def activate(self) -> None:
        if self.action_type == ActionType.Push:
            rule_sections: List[str] = self.rule.split(",")
            finished_rules: List[str] = []
            for rule_section in rule_sections:
                n = Node(
                    None,
                    self.node.grammar,
                    0,
                    _ParsedRuleSection(node_type=NodeType.Raw, raw=rule_section),
                )
                n.expand()
                finished_rules.append(n.finished_text)
            self.node.grammar.push_rules(self.target, finished_rules)

        elif self.action_type == ActionType.Pop:
            self.node.grammar.pop_rules(self.target)

        elif self.action_type == ActionType.Flatten:
            self.node.grammar.flatten(self.target, True)

    def to_text(self) -> str:
        raise NotImplementedError()  # FIXME


class RuleSet:
    """A collection of strings that a symbol can expand to."""

    __slots__ = "grammar", "default_uses", "default_rules"

    grammar: Grammar
    """The grammar a rule set belongs to."""

    default_uses: List[Any]
    """Cached records of when this rule was used."""

    default_rules: List[str]
    """Strings that the rule could evaluate to."""

    def __init__(self, grammar: Grammar, raw: Union[str, List[str]]) -> None:
        self.grammar = grammar
        self.default_uses = []
        self.default_rules = []

        if isinstance(raw, list):
            self.default_rules = raw
        else:
            self.default_rules = [raw]

    def select_rule(self) -> str:
        """Randomly select a rule."""
        # in kate's code there's a bunch of stuff for different methods of
        # selecting a rule, none of which seem to be implemented yet! so for
        # now I'm just going to ...
        return self.grammar.rng.choice(self.default_rules)

    def clear_state(self) -> None:
        """Clear all times this rule was used."""
        self.default_uses = []


@dataclass
class _SymbolUseEntry:
    node: Node


class Symbol:
    """A replacement symbol in a Grammar"""

    __slots__ = "grammar", "key", "raw_rules", "base_rules", "stack", "uses"

    grammar: Grammar
    """The Grammar a symbol belongs to."""

    key: str
    """The string a symbol is mapped to."""

    raw_rules: Union[str, List[str]]
    """The raw substitution rules."""

    base_rules: RuleSet
    """The default set of substitution rules."""

    stack: List[RuleSet]
    """A stack of RuleSets to evaluate."""

    uses: List[_SymbolUseEntry]
    """Records of nodes a symbol is used in a Grammar."""

    def __init__(
        self, grammar: Grammar, key: str, raw_rules: Union[str, List[str]]
    ) -> None:
        self.grammar = grammar
        self.key = key
        self.raw_rules = raw_rules
        self.base_rules = RuleSet(grammar, raw_rules)
        self.stack = [self.base_rules]
        self.uses = []
        self.clear_state()

    def clear_state(self) -> None:
        """Clear cached symbol state."""
        self.stack = [self.base_rules]
        self.uses = []
        self.base_rules.clear_state()

    def push_rules(self, raw_rules: Union[str, List[str]]) -> None:
        """Add new rules to a symbol."""
        rules = RuleSet(self.grammar, raw_rules)
        self.stack.append(rules)

    def pop_rules(self):
        """Remove the most recently pushed rules."""
        self.stack.pop()

    def select_rule(self, node: Node, errors: List[str]) -> str:
        """Select a rule for a given node."""
        self.uses.append(_SymbolUseEntry(node))
        if len(self.stack) == 0:
            errors.append(
                "The rule stack for '" + self.key + "' is empty, too many pops?"
            )
        return self.stack[-1].select_rule()

    def get_active_rules(self) -> Optional[str]:
        """Get a rule from the top of the rule stack"""
        if len(self.stack) == 0:
            return None
        return self.stack[-1].select_rule()


class Grammar:
    """A Tracery Grammar containing symbols and rules that can be expanded/flattened."""

    __slots__ = (
        "rng",
        "symbols",
        "modifiers",
        "errors",
        "raw",
        "sub_grammars",
        "settings",
    )

    rng: random.Random
    """Random number generator for rule selection."""

    symbols: Dict[str, Symbol]
    """Symbols in loaded in a grammar."""

    modifiers: Dict[str, Modifier]
    """Loaded modifier functions."""

    errors: List[str]
    """Errors accumulated during grammar expansion."""

    raw: Dict[str, Union[str, List[str]]]
    """The raw data used to create the grammar."""

    sub_grammars: List[Grammar]
    """Child grammars."""

    settings: Dict[str, Any]
    """Keyword settings."""

    def __init__(
        self,
        raw: Union[
            Dict[str, Union[str, List[str]]], Dict[str, str], Dict[str, List[str]]
        ],
        settings: Optional[Dict[str, Any]] = None,
        rng: Optional[random.Random] = None,
        modifiers: Optional[Dict[str, Modifier]] = None,
    ) -> None:
        self.raw = {k: v for k, v in raw.items()}
        self.rng = rng if rng is not None else random.Random()
        self.modifiers = modifiers if modifiers is not None else {}
        self.errors = []
        self.symbols = {k: Symbol(self, k, v) for k, v in raw.items()}
        self.sub_grammars = []  # Not Used
        self.settings = settings if settings is not None else {}  # Not Used

    def clear_state(self) -> None:
        """Clear all cached states."""
        for val in self.symbols.values():
            val.clear_state()

    def add_modifiers(self, modifiers: Dict[str, Modifier]) -> None:
        """Load modifier functions."""
        for key, modifier in modifiers.items():
            self.modifiers[key] = modifier

    def create_root(self, rule: str) -> Node:
        return Node(None, self, 0, _ParsedRuleSection(node_type=NodeType.Raw, raw=rule))

    def expand(self, rule: str, allow_escape_chars: bool = False) -> Node:
        root = self.create_root(rule)
        root.expand()
        if not allow_escape_chars:
            root.clear_escape_chars()
        self.errors.extend(root.errors)
        return root

    def flatten(self, rule: str, allow_escape_chars: bool = False) -> str:
        """Expand a rule using the Grammar."""
        root = self.expand(rule, allow_escape_chars)
        return root.finished_text

    def push_rules(self, key: str, raw_rules: Union[str, List[str]]) -> None:
        if key not in self.symbols:
            self.symbols[key] = Symbol(self, key, raw_rules)
        else:
            self.symbols[key].push_rules(raw_rules)

    def pop_rules(self, key: str) -> None:
        if key not in self.symbols:
            self.errors.append("Can't pop: no symbol for key " + key)
        else:
            self.symbols[key].pop_rules()

    def select_rule(self, key: Optional[str], node: Node, errors: List[str]) -> str:
        if key in self.symbols:
            return self.symbols[key].select_rule(node, errors)
        else:
            if key is None:
                key = str(None)
            self.errors.append("No symbol for " + key)
            return "((" + key + "))"


class Modifier(Protocol):
    """A Tracery modifier function."""

    def __call__(self, text: str, *params: str) -> str:
        raise NotImplementedError()


@dataclass
class _ParsedTagOutput:
    symbol: Optional[str] = None
    pre_actions: List[_ParsedRuleSection] = field(default_factory=list)
    post_actions: List[_ParsedRuleSection] = field(default_factory=list)
    modifiers: List[str] = field(default_factory=list)


def parse_tag(tag_contents: str) -> _ParsedTagOutput:
    """
    returns a dictionary with 'symbol', 'modifiers', 'preactions',
    'postactions'
    """
    parsed: _ParsedTagOutput = _ParsedTagOutput()

    sections, _ = parse(tag_contents)
    symbol_section = None
    for section in sections:
        if section.node_type == NodeType.PlainText:
            if symbol_section is None:
                symbol_section = section.raw
            else:
                raise Exception("multiple main sections in " + tag_contents)
        else:
            parsed.pre_actions.append(section)

    if symbol_section is not None:
        components = symbol_section.split(".")
        parsed.symbol = components[0]
        parsed.modifiers = components[1:]

    return parsed


@dataclass
class _ParsedRuleSection:
    node_type: NodeType = NodeType.Raw
    raw: str = ""
    pre_actions: List[Action] = field(default_factory=list)


def parse(rule: Optional[str]) -> Tuple[List[_ParsedRuleSection], List[str]]:
    depth = 0
    in_tag = False
    sections: List[_ParsedRuleSection] = list()
    escaped = False
    errors: List[str] = []
    start = 0
    escaped_substring = ""
    last_escaped_char = None

    if rule is None:
        return sections, errors

    def create_section(start: int, end: int, type_: NodeType):
        if end - start < 1:
            if type_ == NodeType.Tag:
                errors.append(str(start) + ": empty tag")
            elif type_ == NodeType.Action:
                errors.append(str(start) + ": empty action")
        raw_substring = None
        if last_escaped_char is not None:
            raw_substring = escaped_substring + "\\" + rule[last_escaped_char + 1 : end]
        else:
            raw_substring = rule[start:end]
        sections.append(_ParsedRuleSection(node_type=type_, raw=raw_substring))

    for i, c in enumerate(rule):
        if not escaped:
            if c == "[":
                if depth == 0 and not in_tag:
                    if start < i:
                        create_section(start, i, NodeType.PlainText)
                        last_escaped_char = None
                        escaped_substring = ""
                    start = i + 1
                depth += 1
            elif c == "]":
                depth -= 1
                if depth == 0 and not in_tag:
                    create_section(start, i, NodeType.Action)
                    last_escaped_char = None
                    escaped_substring = ""
                    start = i + 1
            elif c == "#":
                if depth == 0:
                    if in_tag:
                        create_section(start, i, NodeType.Tag)
                        last_escaped_char = None
                        escaped_substring = ""
                        start = i + 1
                    else:
                        if start < i:
                            create_section(start, i, NodeType.PlainText)
                            last_escaped_char = None
                            escaped_substring = ""
                        start = i + 1
                    in_tag = not in_tag
            elif c == "\\":
                escaped = True
                escaped_substring = escaped_substring + rule[start:i]
                start = i + 1
                last_escaped_char = i
        else:
            escaped = False
    if start < len(rule):
        create_section(start, len(rule), NodeType.PlainText)
        last_escaped_char = None
        escaped_substring = ""

    if in_tag:
        errors.append("unclosed tag")
    if depth > 0:
        errors.append("too many [")
    if depth < 0:
        errors.append("too many ]")

    sections = [
        s
        for s in sections
        if not (s.node_type == NodeType.PlainText and len(s.raw) == 0)
    ]

    return sections, errors



"""Tracery modifiers."""


def replace(text: str, *params: str) -> str:
    """Replace text with a given string.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    return text.replace(params[0], params[1])


def capitalize_all(text: str, *params: str) -> str:
    """Convert text to title case, capitalizing the first letter of all words.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    return text.title()


def capitalize(text: str, *params: str) -> str:
    """Capitalize the first letter.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    return text[0].upper() + text[1:]


def a(text: str, *params: str) -> str:
    """Add proper (a/an) article before noun.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    if len(text) > 0:
        if text[0] in "uU":
            if len(text) > 2:
                if text[2] in "iI":
                    return "a " + text
        if text[0] in "aeiouAEIOU":
            return "an " + text
    return "a " + text


def first_s(text: str, *params: str) -> str:
    """Pluralize first word of text.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    text2 = text.split(" ")
    return " ".join([s(text2[0])] + text2[1:])


def s(text: str, *params: str) -> str:
    """Pluralize text.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    if text[-1] in "shxSHX":
        return text + "es"
    elif text[-1] in "yY":
        if text[-2] not in "aeiouAEIOU":
            return text[:-1] + "ies"
        else:
            return text + "s"
    else:
        return text + "s"


def ed(text: str, *params: str) -> str:
    """Convert text to past-tense.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    if text[-1] in "eE":
        return text + "d"

    elif text[-1] in "yY":
        if text[-2] not in "aeiouAEIOU":
            return text[:-1] + "ied"

    return text + "ed"


def uppercase(text: str, *params: str) -> str:
    """Convert text to uppercase.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    return text.upper()


def lowercase(text: str, *params: str) -> str:
    """Convert text to lowercase.

    Parameters
    ----------
    text
        The text to modify
    *params
        Additional parameters for the modifier

    Returns
    -------
    str
        A modified version of the text
    """
    return text.lower()


base_english = {
    "replace": replace,
    "capitalizeAll": capitalize_all,
    "capitalize": capitalize,
    "a": a,
    "firstS": first_s,
    "s": s,
    "ed": ed,
    "uppercase": uppercase,
    "lowercase": lowercase,
}
"""Modifier functions packaged in a dictionary similar to JS implementation."""



#!/usr/bin/env python3

"""
Unit tests for pytracery
"""
from __future__ import print_function, unicode_literals
from typing import Optional
import unittest

import tracery
from tracery.modifiers import base_english


test_grammar = {
    'deepHash': ["\\#00FF00", "\\#FF00FF"],
    'deeperHash': ["#deepHash#"],
    'animal': ["bear", "cat", "dog", "fox", "giraffe", "hippopotamus"],
    'mood': ["quiet", "morose", "gleeful", "happy", "bemused", "clever",
             "jovial", "vexatious", "curious", "anxious", "obtuse", "serene",
             "demure"],
    'nonrecursiveStory': ["The #pet# went to the beach."],
    'recursiveStory': [
        "The #pet# opened a book about[pet:#mood# #animal#] #pet.a#. "
        "#story#[pet:POP] The #pet# closed the book."],
    'svgColor': ["rgb(120,180,120)", "rgb(240,140,40)", "rgb(150,45,55)",
                 "rgb(150,145,125)", "rgb(220,215,195)", "rgb(120,250,190)"],
    'svgStyle': ['style="fill:#svgColor#;stroke-width:3;stroke:#svgColor#"'],
    "name": ["Cheri", "Fox", "Morgana", "Jedoo", "Brick", "Shadow", "Krox",
             "Urga", "Zelph"],
    "story": ["#hero.capitalize# was a great #occupation#, and this song "
              "tells of #heroTheir# adventure. #hero.capitalize# #didStuff#, "
              "then #heroThey# #didStuff#, then #heroThey# went home "
              "to read a book."],
    "monster": ["dragon", "ogre", "witch", "wizard", "goblin", "golem",
                "giant", "sphinx", "warlord"],
    "setPronouns": [
        "[heroThey:they][heroThem:them][heroTheir:their][heroTheirs:theirs]",
        "[heroThey:she][heroThem:her][heroTheir:her][heroTheirs:hers]",
        "[heroThey:he][heroThem:him][heroTheir:his][heroTheirs:his]"],
    "setOccupation": [
        "[occupation:baker][didStuff:baked bread,decorated cupcakes,"
        "folded dough,made croissants,iced a cake]",
        "[occupation:warrior][didStuff:fought #monster.a#,saved a village "
        "from #monster.a#,battled #monster.a#,defeated #monster.a#]"],
    "origin": ["#[#setPronouns#][#setOccupation#][hero:#name#]story#"]
    }


class TestPytracery(unittest.TestCase):
    """Common setUp and helpers"""

    def setUp(self):
        self.grammar = tracery.Grammar(test_grammar)
        self.grammar.add_modifiers(base_english)

    def assert_starts_with(self, a: str, b: str, msg: Optional[str]=None):
        self.assertTrue(
            a.startswith(b),
            msg or "{} does not start with {}".format(a, b))

    def assert_ends_with(self, a: str, b: str, msg:  Optional[str]=None):
        self.assertTrue(
            a.endswith(b),
            msg or "{} does not end with {}".format(a, b))


class TestBasics(TestPytracery):

    def test_plaintext_short(self):
        # Arrange
        src = "a"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "a")
        self.assertEqual(self.grammar.errors, [])

    def test_plaintext_long(self):
        # Arrange
        src = ("Emma Woodhouse, handsome, clever, and rich, with a "
               "comfortable home and happy disposition, seemed to unite some "
               "of the best blessings of existence; and had lived nearly "
               "twenty-one years in the world with very little to distress or "
               "vex her.")

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(
            ret,
            "Emma Woodhouse, handsome, clever, and rich, with a "
            "comfortable home and happy disposition, seemed to unite some "
            "of the best blessings of existence; and had lived nearly "
            "twenty-one years in the world with very little to distress or "
            "vex her.")
        self.assertEqual(self.grammar.errors, [])


class TestEscapeChars(TestPytracery):

    def test_escape_character(self):
        # Arrange
        src = "\\#escape hash\\# and escape slash\\\\"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "#escape hash# and escape slash\\")
        self.assertEqual(self.grammar.errors, [])

    def test_escape_deep(self):
        # Arrange
        src = "#deepHash# [myColor:#deeperHash#] #myColor#"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertIn(
            ret,
            ["#00FF00  #00FF00",
             "#00FF00  #FF00FF",
             "#FF00FF  #00FF00",
             "#FF00FF  #FF00FF"])
        self.assertEqual(self.grammar.errors, [])

    def test_escape_quotes(self):
        # Arrange
        src = "\"test\" and \'test\'"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "\"test\" and 'test'")
        self.assertEqual(self.grammar.errors, [])

    def test_escape_brackets(self):
        # Arrange
        src = "\\[\\]"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "[]")
        self.assertEqual(self.grammar.errors, [])

    def test_escape_hash(self):
        # Arrange
        src = "\\#"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "#")
        self.assertEqual(self.grammar.errors, [])

    def test_unescape_char_slash(self):
        # Arrange
        src = "\\\\"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "\\")
        self.assertEqual(self.grammar.errors, [])

    def test_escape_melange1(self):
        # Arrange
        src = "An action can have inner tags: \[key:\#rule\#\]"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "An action can have inner tags: [key:#rule#]")
        self.assertEqual(self.grammar.errors, [])

    def test_escape_melange2(self):
        # Arrange
        src = ("A tag can have inner actions: "
               "\"\\#\\[myName:\\#name\\#\\]story\\[myName:POP\\]\\#\"")

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(
            ret,
            "A tag can have inner actions: "
            "\"#[myName:#name#]story[myName:POP]#\"")
        self.assertEqual(self.grammar.errors, [])


class TestWebSpecifics(TestPytracery):

    def test_emoji(self):
        # Arrange
        src = "💻🐋🌙🏄🍻"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(
            ret,
            "\U0001f4bb\U0001f40b\U0001f319\U0001f3c4\U0001f37b")
        self.assertEqual(self.grammar.errors, [])

    def test_unicode(self):
        # Arrange
        src = "&\\#x2665; &\\#x2614; &\\#9749; &\\#x2665;"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "&#x2665; &#x2614; &#9749; &#x2665;")
        self.assertEqual(self.grammar.errors, [])

    def test_unicode_rule_not_found(self):
        # Arrange
        rules = {
            'origin': '#пправило#',
            'rule not found': 'something',
        }
        grammar_rule_not_found = tracery.Grammar(rules)

        # Act
        ret = grammar_rule_not_found.flatten("#origin#")

        # Assert
        self.assertEqual(ret, "((пправило))")
        self.assertEqual(grammar_rule_not_found.errors,
                         ["No symbol for пправило"])

    def test_svg(self):
        # Arrange
        src = ('<svg width="100" height="70">'
               '<rect x="0" y="0" width="100" height="100" #svgStyle#/> '
               '<rect x="20" y="10" width="40" height="30" #svgStyle#/>'
               '</svg>')

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assert_starts_with(
            ret,
            '<svg width="100" height="70">'
            '<rect x="0" y="0" width="100" height="100" style="fill:rgb(')
        self.assertIn('/> <rect x="20" y="10" width="40" height="30" ', ret)
        self.assert_ends_with(ret, '"/></svg>')
        self.assertEqual(self.grammar.errors, [])


class TestPush(TestPytracery):

    def test_push_basic(self):
        # Arrange
        src = "[pet:#animal#]You have a #pet#. Your #pet# is #mood#."

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assert_starts_with(ret, "You have a ")
        self.assertIn(". Your ", ret)
        self.assert_ends_with(ret, ".")
        self.assertEqual(self.grammar.errors, [])

    def test_push_pop(self):
        # Arrange
        src = ("[pet:#animal#]You have a #pet#. [pet:#animal#]Pet:#pet# "
               "[pet:POP]Pet:#pet#")

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assert_starts_with(ret, "You have a ")
        self.assertIn(". Pet:", ret)
        self.assertEqual(self.grammar.errors, [])

    def test_tag_action(self):
        # Arrange
        src = "#[pet:#animal#]nonrecursiveStory# post:#pet#"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assert_starts_with(ret, "The ")
        self.assertIn(" went to the beach. post:", ret)
        self.assertEqual(self.grammar.errors, [])

    def test_test_complex_grammar(self):
        # Arrange
        src = "#origin#"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertIn(", and this song tells of ", ret)
        self.assertEqual(self.grammar.errors, [])

    def test_missing_modifier(self):
        # Arrange
        src = "#animal.foo#"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assert_ends_with(ret, "((.foo))")
        self.assertEqual(self.grammar.errors, [])

    def test_modifier_with_params(self):
        # Arrange
        src = ("[pet:#animal#]#nonrecursiveStory# -> #nonrecursiveStory"
               ".replace(beach,mall)#")

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertIn(" went to the beach. -> ", ret)
        self.assert_ends_with(ret, " went to the mall.")
        self.assertEqual(self.grammar.errors, [])

    def test_recursive_push(self):
        # Arrange
        src = "[pet:#animal#]#recursiveStory#"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertIn(" opened a book about a", ret)
        self.assert_ends_with(ret, " closed the book.")
        self.assertEqual(
            self.grammar.errors,
            ['No symbol for hero', 'No symbol for occupation',
             'No symbol for heroTheir', 'No symbol for hero',
             'No symbol for didStuff', 'No symbol for heroThey',
             'No symbol for didStuff', 'No symbol for heroThey'])


class TestStrings(TestPytracery):

    def test_string_input(self):
        """The grammar should work properly if the input is a string rather
           than a list"""
        # This errored in Py 3.5, taken straight from the README
        rules = {
            'origin': '#hello.capitalize#, #location#!',
            'hello': ['hello'],
            'location': ['world']
        }

        grammar = tracery.Grammar(rules)
        grammar.add_modifiers(base_english)
        self.assertEqual("Hello, world!", grammar.flatten("#origin#"))

    def test_upper_and_lowercase(self):
        rules = {
            'origin': '#hello.lowercase#, #location.uppercase#!',
            'hello': ['Hello'],
            'location': ['world']
        }

        grammar = tracery.Grammar(rules)
        grammar.add_modifiers(base_english)
        self.assertEqual("hello, WORLD!", grammar.flatten("#origin#"))


class TestErrors(TestPytracery):

    def test_plaintext_short(self):
        # Arrange
        src = "a"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "a")
        self.assertEqual(self.grammar.errors, [])

    def test_unmatched_hash(self):
        # Arrange
        src = "#unmatched"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "unmatched")
        self.assertEqual(self.grammar.errors, ["unclosed tag"])

    def test_missing_symbol(self):
        # Arrange
        src = "#unicorns#"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "((unicorns))")
        self.assertEqual(self.grammar.errors, ["No symbol for unicorns"])

    def test_missing_right_bracket(self):
        # Arrange
        src = "[pet:unicorn"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "pet:unicorn")
        self.assertEqual(self.grammar.errors, ["too many ["])

    def test_missing_left_bracket(self):
        # Arrange
        src = "pet:unicorn]"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "pet:unicorn]")
        self.assertEqual(self.grammar.errors, ["too many ]"])

    def test_just_a_lot_of_brackets(self):
        # Arrange
        src = "[][]][][][[[]]][[]]]]"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "][][][][]]]")
        self.assertEqual(
            self.grammar.errors,
            ["1: empty action", "1: empty action", "3: empty action",
             "17: empty action", "too many ]"])

    def test_bracket_tag_melange(self):
        # Arrange
        src = "[][#]][][##][[[##]]][#[]]]]"

        # Act
        ret = self.grammar.flatten(src)

        # Assert
        self.assertEqual(ret, "][][((None))][][[]]]]")
        self.assertEqual(
            self.grammar.errors,
            ['unclosed tag', 'No symbol for None', 'No symbol for None',
             '1: empty tag', '1: empty action', '10: empty tag',
             'unclosed tag', 'too many ]'])


if __name__ == "__main__":
    unittest.main()

# End of file



#!/usr/bin/env python3
"""Unit tests for pytracery."""

from __future__ import annotations

import unittest

from tracery import modifiers


class TestModifiers(unittest.TestCase):
    def test_replace(self):
        # Arrange
        text = "a big ship"

        # Act
        output = modifiers.replace(text, "ship", "shop")

        # Assert
        self.assertEqual(output, "a big shop")

    def test_capitalize_all(self):
        # Arrange
        text = "abc def"

        # Act
        output = modifiers.capitalize_all(text)

        # Assert
        self.assertEqual(output, "Abc Def")

    def test_capitalize_(self):
        # Arrange
        text = "abc def"

        # Act
        output = modifiers.capitalize(text)

        # Assert
        self.assertEqual(output, "Abc def")

    def test_a_consonant(self):
        # Arrange
        text = "house"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "a house")

    def test_a_consonant_uppercase(self):
        # Arrange
        text = "HOUSE"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "a HOUSE")

    def test_a_u_something_i(self):
        # Arrange
        text = "unicorn"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "a unicorn")

    def test_a_u_something_i_uppercase(self):
        # Arrange
        text = "UNICORN"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "a UNICORN")

    def test_a_u_not_i(self):
        # Arrange
        text = "underdog"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "an underdog")

    def test_a_u_not_i_uppercase(self):
        # Arrange
        text = "UNDERDOG"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "an UNDERDOG")

    def test_a_vowel(self):
        # Arrange
        text = "elephant"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "an elephant")

    def test_a_vowel_uppercase(self):
        # Arrange
        text = "ELEPHANT"

        # Act
        output = modifiers.a(text)

        # Assert
        self.assertEqual(output, "an ELEPHANT")

    def test_first_s(self):
        # Arrange
        text = "elephant in a phonebox"

        # Act
        output = modifiers.first_s(text)

        # Assert
        self.assertEqual(output, "elephants in a phonebox")

    def test_first_s_uppercase(self):
        # Arrange
        text = "ELEPHANT IN A PHONEBOX"

        # Act
        output = modifiers.first_s(text)

        # Assert
        self.assertEqual(output, "ELEPHANTs IN A PHONEBOX")

    def test_s_ends_in_x(self):
        # Arrange
        text = "box"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "boxes")

    def test_s_ends_in_x_uppercase(self):
        # Arrange
        text = "BOX"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "BOXes")

    def test_s_ends_in_non_s(self):
        # Arrange
        text = "goat"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "goats")

    def test_s_ends_in_non_s_uppercase(self):
        # Arrange
        text = "GOAT"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "GOATs")

    def test_s_ends_in_vowel_y(self):
        # Arrange
        text = "monkey"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "monkeys")

    def test_s_ends_in_vowel_y_uppercase(self):
        # Arrange
        text = "MONKEY"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "MONKEYs")

    def test_s_ends_in_y_but_not_vowel_y(self):
        # Arrange
        text = "telly"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "tellies")

    def test_s_ends_in_y_but_not_vowel_y_uppercase(self):
        # Arrange
        text = "TELLY"

        # Act
        output = modifiers.s(text)

        # Assert
        self.assertEqual(output, "TELLies")

    def test_ed_ends_in_e(self):
        # Arrange
        text = "glide"

        # Act
        output = modifiers.ed(text)

        # Assert
        self.assertEqual(output, "glided")

    def test_ed_ends_in_e_uppercase(self):
        # Arrange
        text = "GLIDE"

        # Act
        output = modifiers.ed(text)

        # Assert
        self.assertEqual(output, "GLIDEd")

    def test_ed_ends_in_y_but_not_vowel_y(self):
        # Arrange
        text = "shimmy"

        # Act
        output = modifiers.ed(text)

        # Assert
        self.assertEqual(output, "shimmied")

    def test_ed_ends_in_y_but_not_vowel_y_uppercase(self):
        # Arrange
        text = "SHIMMY"

        # Act
        output = modifiers.ed(text)

        # Assert
        self.assertEqual(output, "SHIMMied")

    def test_ed_ends_in_non_e_and_non_y(self):
        # Arrange
        text = "jump"

        # Act
        output = modifiers.ed(text)

        # Assert
        self.assertEqual(output, "jumped")

    def test_ed_ends_in_non_e_and_non_y_uppercase(self):
        # Arrange
        text = "JUMP"

        # Act
        output = modifiers.ed(text)

        # Assert
        self.assertEqual(output, "JUMPed")

    def test_uppercase(self):
        # Arrange
        text = "jump up"

        # Act
        output = modifiers.uppercase(text)

        # Assert
        self.assertEqual(output, "JUMP UP")

    def test_uppercase_uppercase(self):
        # Arrange
        text = "JUMP UP"

        # Act
        output = modifiers.uppercase(text)

        # Assert
        self.assertEqual(output, "JUMP UP")

    def test_lowercase(self):
        # Arrange
        text = "GET DOWN"

        # Act
        output = modifiers.lowercase(text)

        # Assert
        self.assertEqual(output, "get down")

    def test_lowercase_lowercase(self):
        # Arrange
        text = "get down"

        # Act
        output = modifiers.lowercase(text)

        # Assert
        self.assertEqual(output, "get down")


if __name__ == "__main__":
    unittest.main()
